#include "find_t_visitor.hpp"
#include "split.hpp"

#include <type_traits>
#include <string>
#include <regex>

namespace duck::detail {
template <typename T>
FindTVisitor<T>::FindTVisitor (std::string_view search_path) {
	auto split = dincore::split(search_path, '/', true, true);
	m_search.reserve(split.size());
	for (const auto& itm : split) {
		m_search.emplace_back(itm.data(), itm.size());
	}
}

template <typename T>
void FindTVisitor<T>::reset() {
	m_search_error = false;
	m_search_match_count = 0;
}

template <typename T>
bool FindTVisitor<T>::has_error() const noexcept {
	return m_search_error;
}

template <typename T>
T FindTVisitor<T>::operator() (TorrentIntType value) {
	if constexpr (std::is_same<TorrentIntType, T>::value)
		return value;
	m_search_error = true;
	return {};
}

template <typename T>
T FindTVisitor<T>::operator() (const TorrentStringType& value) {
	if constexpr (std::is_same<TorrentStringType, T>::value) {
		if (is_path_found())
			return value;
	}
	m_search_error = true;
	return {};
}

template <typename T>
T FindTVisitor<T>::operator() (const std::vector<TorrentValue>& values) {
	if (is_path_found())
		return {};

	if constexpr (std::is_same<TorrentIntType, T>::value) {
		if (m_search[m_search_match_count] == "[[size]]")
			return static_cast<TorrentIntType>(values.size());
	}

	std::regex reg_index(R"(^\[\[(\d+)\]\]$)");
	std::smatch match;
	const bool matched = std::regex_match(m_search[m_search_match_count], match, reg_index);
	if (matched) {
		++m_search_match_count;
		const auto index = static_cast<std::size_t>(std::stoul(match[1]));
		if (index < values.size())
			return boost::apply_visitor(*this, values[index]);
	}
	return {};
}

template <typename T>
T FindTVisitor<T>::operator() (const std::map<TorrentStringType, TorrentValue>& value) {
	if (is_path_found()) {
		m_search_error = true;
		return {};
	}

	auto it_found = value.find(m_search[m_search_match_count]);
	if (value.end() == it_found)
		return {};

	++m_search_match_count;
	return boost::apply_visitor(*this, it_found->second);
}

template <typename T>
bool FindTVisitor<T>::is_path_found() const {
	return m_search.size() == m_search_match_count;
}

template class FindTVisitor<TorrentStringType>;
template class FindTVisitor<TorrentIntType>;
} //namespace duck::detail
